# Evaluation module for text processing course 2016
# @author Denis Turdakov (turdakov@ispras.ru)

from collections import Counter  # used as multiset
import json
import logging

from sklearn.model_selection import KFold
from statistics import mean
from temp.template_2016 import Solution


class Evaluation:
    def __init__(self, solution, path_to_training, path_to_testing=None, logger_level="DEBUG"):
        """Constructor. It has two modes:
        * if path_to_testing corpus is not None, it used in normal mode for evaluation
        * otherwise it could be used only for cross-validation using function with same name
        :param solution: instance of solution (Solution()) (mandatory)
        :param path_to_training: path to training corpus (mandatory)
        :param path_to_testing: path to testing corpus (optional, None value is by default)
        """
        self.solution = solution

        logging_format = '%(asctime)-15s %(message)s'
        logging.basicConfig(format=logging_format)
        self.logger = logging.getLogger("evaluation")
        self.logger.setLevel(logger_level)

        with open(path_to_training) as f:
            training_corpus = json.load(f)

        if path_to_testing:
            # training of the model
            self.logger.debug("Starting models training")
            self.solution.train(training_corpus)
            self.logger.debug("Training is finished")
            with open(path_to_testing) as f:
                self.testing_corpus = json.load(f)
        else:
            self.training_corpus = training_corpus

    def eval_terms(self):
        """This function evaluates term detector
        Micro averaging between documents is used
        :return: Average F1-score of term detector
        """
        self.logger.debug("Starting evaluation of term detector")
        answer_num = 0
        values_num = 0
        correct = 0
        for cnt, doc in enumerate(self.testing_corpus):
            self.logger.debug("Terms of document #{0} processing started".format(cnt))
            answers = set(self.solution.get_terms(doc["message"]["text"]))
            self.logger.debug("Terms of document #{0} processing finished".format(cnt))
            values = set([(x["start"], x["end"]) for x in doc["message"]["anns"]])
            answer_num += len(answers)
            values_num += len(values)
            correct += len(answers.intersection(values))
        precision = correct / answer_num
        recall = correct / values_num
        f1 = 2 * precision * recall / (precision + recall)
        self.logger.info("Precision: {0}, recall: {1}, F1-score: {2}".format(precision, recall, f1))
        self.logger.debug("Evaluation of term detector is finished")
        return f1

    def eval_meanings(self):
        """Evaluation of disambiguation
        Micro averaging between documents is used
        :return: Average multi-class accuracy of disambiguation
        """

        def compute_sum_weights(answers, expected):
            sum_weights = 0
            for k, v in answers.items():
                actual = Counter(v)
                exact, inexact = expected.get(k, (Counter([]), Counter([])))
                exact_and_actual = sum((exact & actual).values())
                inexact_and_actual = sum((inexact & actual).values())
                exact_or_actual = sum((exact | actual).values())
                inexact_or_missed = sum((inexact | (actual - exact)).values())
                sum_weights += exact_and_actual / exact_or_actual + \
                               (inexact_and_actual / (sum(exact.values()) + 2) / inexact_or_missed) \
                    if inexact else exact_and_actual / exact_or_actual
            return sum_weights

        self.logger.debug("Starting evaluation of disambiguation")
        sum_weights = 0
        terms_counter = 0
        for cnt, doc in enumerate(self.testing_corpus):
            terms = set([(x["start"], x["end"]) for x in doc["message"]["anns"]])
            self.logger.debug("Meanings of document #{0} processing started".format(cnt))
            answers = self.solution.get_meanings(doc["message"]["text"], terms)
            self.logger.debug("Meanings of document #{0} processing finished".format(cnt))
            expected = {(x['start'], x['end']): (Counter([y["id"] for y in x['objects'] if y['isPrimary']]),
                                                 Counter([y["id"] for y in x['objects'] if not y['isPrimary']]))
                        for x in doc["message"]["anns"]}
            sum_weights += compute_sum_weights(answers, expected)
            terms_counter += len(answers)
        q = sum_weights / terms_counter if terms_counter > 0 else 0
        self.logger.info("Accuracy of disambiguation: {0}".format(q))
        self.logger.debug("Evaluation of disambiguation is finished")
        return q

    def cross_validation(self, num_folds=2):
        scores = []
        for cnt, (train_index, test_index) in enumerate(KFold(num_folds).split(self.training_corpus)):
            self.logger.debug("Cross-validation iteration #{0}".format(cnt))
            training_corpus = [self.training_corpus[i] for i in train_index]
            self.testing_corpus = [self.training_corpus[i] for i in test_index]
            self.logger.debug("Starting models training")
            self.solution.train(training_corpus)
            self.logger.debug("Training is finished")
            scores.append((self.eval_terms(), self.eval_meanings()))
        return scores

# evaluation = Evaluation(Solution(), "../data/test.json", "../data/test.json")
# print("Term detection quality: ", evaluation.eval_terms())
# print("Disambiguation quality: ", evaluation.eval_meanings())

evaluation = Evaluation(Solution(), "../data/test.json", logger_level="INFO")
print("Mean F1 of term detector: {0}, Mean accuracy of disambiguation: {1}"
      .format(*[mean(x) for x in zip(*evaluation.cross_validation())]))